class NvdbError(Exception):
    """Base class for exceptions in this module."""
    pass

class ApiError(NvdbError):
    def __init__(self, message):
        self.message = message


def read_api_error(resp):
    try:
        return resp.json()[0]['message']
    except:
        return 'API seems dead : {}'.format(resp)



from .util import _fetch_data


class Objekt_type(object):
    """ Class for individual nvdb-object types. (Data catalogue) """
    def __init__(self, nvdb, objekt_type, meta=None):
        super(Objekt_type, self).__init__()
        self.nvdb = nvdb
        self.objekt_type = objekt_type
        self.data = None
        self.meta = meta

    def dump(self, format='json'):
        """
        Function for dumping raw API-result for object.

        :param format: Type of data to dump as. json or xml
        :type format: string
        :returns: str
        """
        if format.lower() == 'json':
            if not self.data:
                self.data = _fetch_data(self.nvdb, 'vegobjekttyper/{}'.format(self.objekt_type))
            return self.data
        
        elif format.lower() == 'xml':
            xml_data =_fetch_data(self.nvdb, 'vegobjekttyper/{}.xml'.format(self.objekt_type), format='xml')
            return xml_data

    @property
    def relasjonstyper(self):
        """
        :Attribute: Dict
        """
        if not self.data:
            self.data = _fetch_data(self.nvdb, 'vegobjekttyper/{}'.format(self.objekt_type))
        return self.data['relasjonstyper']

    @property
    def egenskapstyper(self):
        """
        :Attribute: Dict
        """
        if not self.data:
            self.data = _fetch_data(self.nvdb, 'vegobjekttyper/{}'.format(self.objekt_type))
        return self.data['egenskapstyper']
    
    @property
    def styringsparametere(self):
        """
        :Attribute: Dict
        """
        if not self.data:
            self.data = _fetch_data(self.nvdb, 'vegobjekttyper/{}'.format(self.objekt_type))
        return self.data['styringsparametere']

    @property
    def metadata(self):
        """
        :Attribute: Dict
        """
        if self.meta:
            return self.meta
        elif not self.data:
            self.data = _fetch_data(self.nvdb, 'vegobjekttyper/{}'.format(self.objekt_type))
            metadata = self.data.copy()
            del metadata['egenskapstyper']
            del metadata['relasjonstyper']
            del metadata['styringsparametere']
            self.meta = metadata
        return self.meta

    @property
    def barn(self):
        """
        :Attribute: list of :class:`.Objekt_type`
        """
        if not self.data:
            self.data = _fetch_data(self.nvdb, self.baseUrl, 'vegobjekttyper', self.objekt_type)
        realasjoner = self.data['relasjonstyper']
        return [Objekt_type(i['type']['id']) for i in realasjoner['barn']]
    @property
    def foreldre(self):
        """
        :Attribute: list of :class:`.Objekt_type`
        """
        if not self.data:
            self.data = _fetch_data(self.nvdb, self.baseUrl, 'vegobjekttyper', self.objekt_type)
        realasjoner = self.data['relasjonstyper']
        return [Objekt_type(self.nvdb, i['type']['id']) for i in realasjoner['foreldre']]


import requests
from .pnvdb_exceptions import *
class NVDBase(object):
    """Superclass for all models in pvdb"""
    def __init__(self, client, contact):
        self.baseUrl = 'https://www.vegvesen.no/nvdb/api/v2'
        self.headers = {'X-Client': client,'X-Kontaktperson': contact}
        self.srid = ''
        self.antall = 1000



def _fetch_data(nvdb, url_add, payload={}, format='json'):
    url = '{baseUrl}/{url_add}'.format(baseUrl=nvdb.baseUrl, url_add=url_add)
    resp = requests.get(url, params=payload, headers=nvdb.headers)
    #print(resp.headers)
    #print(resp.url)
    data = _check_response(nvdb, resp, format)
    return data

def _check_response(nvdb, resp, format='json'):
    """Function verifes that a 200 code was returned from the API
    and returns the data as Json.
    If a 200 code was not returned, it tries to return the error recived 
    from the API."""
    if resp.status_code == requests.codes.ok and format == 'json':
        return resp.json()
    elif resp.status_code == requests.codes.ok and format == 'xml':
        return resp
    else:
        print(resp.url)
        raise ApiError(read_api_error(resp))


from .util import _fetch_data
from .objekt import Objekt

class Area(object):
    """ Class for area objects. """ 
    def __init__(self, nvdb, area_data):
        super(Area, self).__init__()
        self.nvdb = nvdb
        self.data = area_data

    @property
    def metadata(self):
        """ 
        :Attribute: Dict
        """
        metadata = self.data.copy()
        if 'kartutsnitt' in self.data:
            del metadata['kartutsnitt']
        if 'senterpunkt' in self.data:
            del metadata['senterpunkt']
        if 'vegobjekt' in self.data:
            del metadata['vegobjekt']
        return metadata


    @property
    def kartutsnitt(self):
        """ 
        :Attribute: Well Known Text
        """
        if 'kartutsnitt' in self.data:
            return self.data['kartutsnitt']['wkt']
        else:
            return None
    
    
    @property
    def senterpunkt(self):
        """
        :Attribute: Well Known Text
        """
        if 'senterpunkt' in self.data:
            return self.data['senterpunkt']['wkt']
        else:
            return None
    

    @property
    def objekt(self):
        """
        :Attribute: :class:`.Objekt` of the Area
        """
        objekttype = self.data['vegobjekt']['type']
        nvdb_id = self.data['vegobjekt']['id']
        return Objekt(self.nvdb, objekttype, nvdb_id)


from .util import _fetch_data
from .vegreferanse import Vegreferanse

class Posisjon(object):
    """ Class for connecting coordinates to road refferences """
    def __init__(self, nvdb, payload):
        super(Posisjon, self).__init__()
        self.nvdb = nvdb
        self.data = _fetch_data(self.nvdb, 'posisjon',payload)
        
    
    @property
    def vegreferanse(self):
        """
        :Attribute: :class:`.Vegreferanse`
        """
        return Vegreferanse(self.nvdb, self.data[0]['vegreferanse']['kortform'])



# -*- coding: utf-8 -*-
import requests
from . import models
from .models.util import _fetch_data

class Nvdb(object):
    """ The main class for interfacing with the API.
        
        :param client: Name of client using the API
        :type name: str
        :param contact: Contact information of user of the API
        :returns:  Nvdb Class

        >>> import pnvdb
        >>> nvdb = pnvdb.Nvdb(client='Your-App-Name', contact='Your-contact-information')

        
    """
       
    def __init__(self, client='pnvdb', contact=''):
        self.baseUrl = 'https://www.vegvesen.no/nvdb/api/v2'
        self.headers = {'X-Client': client,'X-Kontaktperson': contact}
        self.srid = ''
        self.antall = 1000



    def status(self):
        """ Method for getting information about the current status of the API 
            
            >>> status = nvdb.status()
            >>> print(status['datakatalog']['versjon'])
            2.10

            :returns: Dict
        """
        return _fetch_data(self, 'status')
    
    def objekt(self, objekt_type, nvdb_id):
        """ Method for creating a spesific nvdb python Objekt

            :param objekt_type: nvdb objekttype id.
            :type name: int
            :param nvdb_id: the unique nvdb id
            :type name: int
            :returns: :class:`.Objekt`

            >>> obj = nvdb.objekt(objekt_type=67, nvdb_id=89204552)
            >>> print(obj.metadata)
            {'versjon': 3, 'type':P {'navn': 'Tunnelløp', 'id': 67}, 'startdato': '2014-01-17', 'sist_modifisert': '2017-10-23 15:15:50'}

        """
        return models.Objekt(self, objekt_type, nvdb_id)


    def objekt_type(self, objekt_type):
        """ Method for creating a spesific nvdb python

            :param objekt_type: nvdb objekttype id.
            :type name: int
            :returns: :class:`.models.models.Objekt_type`

            >>> obj = nvdb.objekt_type(objekt_type=67)
            >>> print(obj.metadata['sosinvdbnavn'])
            Tunnelløp_67
        """
        return models.Objekt_type(self, objekt_type)
    

    def objekt_typer(self):
        """ Returns objekt_type of every avaliable obj type in nvdb
            
            :returns: List of :class:`.models.models.Objekt_type`
            
            >>> obj_types = nvdb.objekt_typer()
            >>> print(obj_types[0].metadata['sosinvdbnavn'])
            Skjerm_3

        """
        data = _fetch_data(self, 'vegobjekttyper')
        objekt_typer = []
        for objekt_type in data:
            objekt_type_id = objekt_type['id']
            objekt_typer.append(models.Objekt_type(self, objekt_type_id, meta=objekt_type))
        return objekt_typer


    def hent(self, objekt_type, payload={}):
        """ Return a generator object that can be itterated over
            to fetch the results of the query.
            
            :param objekt_type: nvdb objekttype id.
            :type payload: dict
            :returns: generator of :class:`.Objekt`
            
            >>> criteria = {'fylke':'2','egenskap':'1820>=20'}
            >>> bomstasjoner = nvdb.hent(45, criteria)
            >>> for bomstasjon in bomstasjoner:
            >>>     print(bomstasjon)
        """
        _payload = payload.copy()
        _payload.update({'antall':self.antall})
        url = 'vegobjekter/{objekt_type}'.format(objekt_type=objekt_type)
        data = _fetch_data(self, url, payload=_payload)
        while True:

            returnert = data['metadata']['returnert']
            if returnert == 0:
                break
            
            _payload.update({'start':data['metadata']['neste']['start']})
            for obj in enumerate(data['objekter']):
                yield models.Objekt(self, objekt_type, obj[1]['id'])
            data = _fetch_data(self, url, _payload)
        

    def vegreferanse(self, vegreferanse):
        """ Return vegreferanse object.
            PS : Only support point refferences

            :param vegreferanse: The road refferences to objectify
            :type vegreferanse: string
            :returns: :class:`.Vegreferanse`

            >>> print(nvdb.vegreferanse('1600Ev6hp12m1000'))
        """
        if type(vegreferanse) == list:
            payload = {'vegreferanser':','.join(vegreferanse)}
            result = _fetch_data(self, 'veg/batch', payload)
            return [Vegreferanse(self, vegref, meta=result[vegref]) for vegref in vegreferanse]
        return models.Vegreferanse(self, vegreferanse)
    

    def posisjon(self, x=None, y=None, lat=None, lon=None):
        """Returns a posisjon object for a given location
            
            :param x: X-coordinate in EUREF89 UTM 33
            :type x: float
            :param y: Y-coordinate in EUREF89 UTM 33
            :type y: float
            :param lat: Lattitude in EUREF89
            :type lat: float
            :param lon: Longitude in EUREF89
            :type lon: float
            :returns: :class:`.Posisjon`

            >>> pos = nvdb.posisjon(x=269815,y=7038165)
            >>> print(pos.vegreferanse)
        """
        if x and y:
            payload = {'nord':y,'ost':x}
        elif lat and lon:
            payload = {'lat':lat,'lon':lon}

        return models.Posisjon(self, payload)
    
   
    def regioner(self):
        """ Returns an models.Area object for all regions
            
            :returns: list of :class:`.models.Area`

            >>> for region in nvdb.regioner():
            >>>     print(region.metadata)
        """
        payload = {'inkluder':'alle'}
        data = _fetch_data(self, 'omrader/regioner', payload)
        return [models.Area(self, models.Area_data) for models.Area_data in data]

    def fylker(self):
        """ Returns an models.Area object for all fylker

            :returns: list of :class:`.models.Area`

            >>> for region in nvdb.regioner():
            >>>     print(region.metadata)
        """
        payload = {'inkluder':'alle'}
        data = _fetch_data(self, 'omrader/fylker', payload)
        return [models.Area(self, models.Area_data) for models.Area_data in data]

    def vegavdelinger(self):
        """ Returns an models.Area object for all vegavdelinger

            :returns: list of :class:`.models.Area`

            >>> for region in nvdb.regioner():
            >>>     print(region.metadata)
        """
        payload = {'inkluder':'alle'}
        data = _fetch_data(self, 'omrader/vegavdelinger', payload)
        return [models.Area(self, models.Area_data) for models.Area_data in data]

    def kommuner(self):
        """ Returns an models.Area object for all kommuner

            :returns: list of :class:`.models.Area`

            >>> for region in nvdb.regioner():
            >>>     print(region.metadata)
        """
        payload = {'inkluder':'alle'}
        data = _fetch_data(self, 'omrader/kommuner', payload)
        return [models.Area(self, models.Area_data) for models.Area_data in data]

    def kontraktsomrader(self):
        """ Returns an models.Area object for all kontraktsomrader

            :returns: list of :class:`.models.Area`

            >>> for region in nvdb.regioner():
            >>>     print(region.metadata)
        """
        payload = {'inkluder':'alle'}
        data = _fetch_data(self, 'omrader/kontraktsomrader', payload)
        return [models.Area(self, models.Area_data) for models.Area_data in data]

    def riksvegruter(self):
        """ Returns an models.Area object for all riksvegruter

            :returns: list of :class:`.models.Area`

            >>> for region in nvdb.regioner():
            >>>     print(region.metadata)
        """
        payload = {'inkluder':'alle'}
        data = _fetch_data(self, 'omrader/riksvegruter', payload)
        return [models.Area(self, models.Area_data) for models.Area_data in data]


from .util import _fetch_data
from .vegreferanse import Vegreferanse

class Objekt(object):
    """ Class for individual nvdb-objects. """
    def __init__(self, nvdb, objekt_type, nvdb_id):
        self.nvdb = nvdb
        self.objekt_type = objekt_type
        self.nvdb_id = nvdb_id
        self.data = None

    @property
    def egengeometri(self):
        """
        :Attribute: Well known text
        """
        if not self.data:
           self.data = _fetch_data(self.nvdb, 'vegobjekter/{}/{}'.format(self.objekt_type, self.nvdb_id))
        if 'geometri' in self.data:
            if self.data['geometri']['egengeometri'] == 'true':
                egengeometri = True
            else:
                egengeometri = False
        return egengeometri

    @property
    def egenskaper(self):
        """
        :Attribute: List of Dict
        """
        if not self.data:
           self.data = _fetch_data(self.nvdb, 'vegobjekter/{}/{}'.format(self.objekt_type, self.nvdb_id))
        if 'egenskaper' in self.data:
            egenskaper = self.data['egenskaper']
        else:
            egenskaper = None
        return egenskaper

            
    @property
    def metadata(self):
        """
        :Attribute: Dict
        """
        if not self.data:
           self.data = _fetch_data(self.nvdb, 'vegobjekter/{}/{}'.format(self.objekt_type, self.nvdb_id))
        if 'metadata' in self.data:
            metadata = self.data['metadata']
        else:
            metadata = None
        return self.data['metadata']


    @property
    def geometri(self):
        """
        :Attribute: Dict
        """
        if not self.data:
           self.data = _fetch_data(self.nvdb, 'vegobjekter/{}/{}'.format(self.objekt_type, self.nvdb_id))
        if 'geometri' in self.data:
            geometri = self.data['geometri']['wkt']
        else:
            geometri = None
        return geometri

    def dump(self, format='json'):
        """
        :Attribute: Dict
        """
        if format.lower() == 'json':
            if not self.data:
                self.data = _fetch_data(self.nvdb, 'vegobjekter/{}/{}'.format(self.objekt_type, self.nvdb_id))
            return self.data
        elif format.lower() == 'xml':
            xml_data = _fetch_data(self.nvdb, 'vegobjekter/{}/{}.xml'.format(self.objekt_type, self.nvdb_id), format='xml')
            return xml_data

    @property
    def foreldre(self):
        """
        :Attribute: List of Objekt
        """
        if not self.data:
           self.data = _fetch_data(self.nvdb, 'vegobjekter/{}/{}'.format(self.objekt_type, self.nvdb_id))
        foreldre = []
        if 'relasjoner' in self.data and 'foreldre' in self.data['relasjoner']:
            for i in self.data['relasjoner']['foreldre']:
                objekt_type = i['type']['id']
                for nvdb_id in i['vegobjekter']:
                    foreldre.append(Objekt(self.nvdb, objekt_type, nvdb_id))
        else:
            foreldre = None
        return foreldre

    @property
    def barn(self):
        """
        :Attribute: List of Objekt

        """
        if not self.data:
           self.data = _fetch_data(self.nvdb, 'vegobjekter/{}/{}'.format(self.objekt_type, self.nvdb_id))
        barn = []
        if 'relasjoner' in self.data and 'barn' in self.data['relasjoner']:
            for i in self.data['relasjoner']['barn']:
                objekt_type = i['type']['id']
                for nvdb_id in i['vegobjekter']:
                    barn.append(Objekt(self.nvdb, objekt_type, nvdb_id))

        else:
            barn = None
        return barn

    @property
    def vegreferanser(self):
        """
        :Attribute: Dict

        """
        if not self.data:
           self.data = _fetch_data(self.nvdb, 'vegobjekter/{}/{}'.format(self.objekt_type, self.nvdb_id))
        vegreferanser = []
        if 'lokasjon' in self.data and 'vegreferanser' in self.data['lokasjon']:
            for i in  self.data['lokasjon']['vegreferanser']:
                vegreferanser.append(Vegreferanse(self.nvdb, i['kortform']))
        else:
            vegreferanser = None
        return vegreferanser


from .objekt import Objekt
from .vegreferanse import Vegreferanse
from .area import Area
from .objekt_type import Objekt_type
from .posisjon import Posisjon
__all__ = ('Objekt','Vegreferanse', 'Area', 'objekt_type', 'posisjon')


from .util import _fetch_data

class Vegreferanse(object):
    """ Class for working with road refferences. """
    def __init__(self, nvdb, vegreferanse, meta=None):
        super(Vegreferanse, self).__init__()
        self.vegreferanse = vegreferanse
        self.nvdb = nvdb
        self.data = meta
       
    @property
    def detaljert(self):
        """
        :Attribute: Dict
        """
        if not self.data:
            self.data = _fetch_data(self.nvdb, 'veg', payload={'vegreferanse':self.vegreferanse})
        return self.data['vegreferanse']
        
    @property
    def veglenke(self):
        """
        :Attribute: Dict
        """
        if not self.data:
           self.data = _fetch_data(self.nvdb, 'veg', payload={'vegreferanse':self.vegreferanse})
        return self.data['veglenke']
    
    @property
    def geometri(self):
        """
        :Attribute: Well known text
        """
        if not self.data:
           self.data = _fetch_data(self.nvdb, 'veg', payload={'vegreferanse':self.vegreferanse})
        return self.data['geometri']['wkt']
    
    def __str__(self):
        return '{}'.format(self.vegreferanse)